#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>

const char* _version_ = "pmem 1.1.2 (http://www.pmem.net/ Daniel Etzold <detzold@gmx.de>)";

#define PLOT_OUTPUT "plot.ps"
#define PLOT_DATA   "plot.dat"
#define PLOT_CMD    "plot.gnu"

struct mem_t {
	int _size;
	int _res;
	int _shared;
	int _text;
	int _shared_libs;
	int _stack;
	int _dirty_pages;
};

int show_process_mem( int argc, char** argv, int optind );
int read_mem( int pid, int* size, int* res, int* shared, int* text, int* shared_libs, int* stack, int* dirty_pages );
int read_mem_t( int pid, struct mem_t* mem );
int show_details( int argc, char** argv, int optind );
int logmem( int argc, char** argv, int optind, int ms );
void help( char* prg );

int _show_mb = 0;
int _show_kb = 0;
int _show_size = 0;
int _show_shared = 0;
char* _gnuplot = 0;

int main( int argc, char** argv ) 
{
	static struct option options[] = {
		{ 0, 0, 0, 0 }
	};
	int optidx = 0;
	int details = 0;
	int loginterval = 100;
	int log = 0;
	int h = 0;
	char c;
	while( 1 ) {
		c = getopt_long( argc, argv, "dvzsmhkli:g:", options, &optidx );
		if( c == -1 ) {
			break;
		}
		switch( c ) {
			case '?':
				return 1;
			case 'z':
				_show_size = 1;
				break;
			case 's':
				_show_shared = 1;
				break;
			case 'm':
				_show_mb = 1;
				break;
			case 'k':
				_show_kb = 1;
				break;
			case 'h':
				h = 1;
				break;
			case 'v':
				printf( "%s\n", _version_ );
				return 0;
			case 'd':
				details = 1;
				break;
			case 'l':
				log = 1;
				break;
			case 'i':
				loginterval = atoi(optarg);
				break;
			case 'g':
				_gnuplot = optarg;
				break;
		}
	}
	if( h == 1 ) {
		help( argv[ 0 ] );
		return 0;
	} else {
		if( optind < argc ) {
			if( details ) {
				return show_details( argc, argv, optind );
			} else if (log) {
				return logmem(argc, argv, optind, loginterval);
			} else {
				return show_process_mem( argc, argv, optind );
			}
		} else {
			fprintf( stderr, "no process id given\n" );
		}
	}
	return 1;
}

void help( char* prg )
{
	printf( "usage: %s [OPTIONS] PIDs\nDetermines the memory usage of a process.\nOptions:\n", prg );
	printf( "  -z  show size (default is resident)\n"
			"  -s  show shared memory\n"
			"  -m  convert to MB\n"
			"  -k  convert to KB\n"
			"  -d  show details\n"
			"  -l  log memory usage of given pids\n"
			"    -i <n>  log interval in milliseconds [default: 100]\n"
			"    -g <f>  write a gnuplot command file into plot.gnu\n"
			"  -h  this help screen\n"
			"  -v  version\n" );
}

int format_mem(int mem) 
{
	if (_show_mb) {
		return mem * getpagesize() / (1024 * 1024);
	} else if(_show_kb) {
		return mem * getpagesize() / 1024;
	}
	return mem * getpagesize();
}

int get_mem(struct mem_t* mem) 
{
	if (_show_size) {
		return mem->_size;
	} else if (_show_shared) {
		return mem->_shared;
	}
	return mem->_res;
}

/* Returns 0 on success, -1 on error. */
int read_mem( int pid, int* size, int* res, int* shared, int* text, int* shared_libs, int* stack, int* dirty_pages ) 
{
	int ret = 0;
	char buf[ 4096 ];
	FILE* f;
	sprintf( buf, "/proc/%d/statm", pid );
	f = fopen( buf, "r" );
	if( f ) {
		fscanf( f, "%d %d %d %d %d %d %d", size, res, shared, text, shared_libs, stack, dirty_pages );
		fclose( f );
	} else {
		ret = -1;
	}
	return ret;
}

int read_mem_t( int pid, struct mem_t* mem) 
{
	int ret = 0;
	char buf[ 4096 ];
	FILE* f;
	sprintf( buf, "/proc/%d/statm", pid );
	f = fopen( buf, "r" );
	if( f ) {
		fscanf( f, "%d %d %d %d %d %d %d", &mem->_size, &mem->_res, &mem->_shared, 
				&mem->_text, &mem->_shared_libs, &mem->_stack, &mem->_dirty_pages );
		fclose( f );
	} else {
		ret = -1;
	}
	return ret;
}

int show_details( int argc, char** argv, int optind )
{
	static char* descr[] = { "total process size:", "resident in physical memory:", "shared with other processes:",
		"loaded executable:", "shared libraries:", "stack:", "dirty pages:" };
	int val[ 7 ];
	int ret = 0;
	int pid;
	int r, i;
	printf("page size: %d\n", getpagesize());
	for( ; optind < argc; ++optind ) {
		pid = atoi( argv[ optind ] );
		r = read_mem( pid, &val[ 0 ], &val[ 1 ], &val[ 2 ], &val[ 3 ], &val[ 4 ], &val[ 5 ], &val[ 6 ] );
		if( ! r ) {
			printf( "pid: %d\n", pid );
			for( i = 0; i < 7; ++i ) {
				ret = format_mem(val[i]);
				printf( "%s %d\n", descr[ i ], ret );
			}
		} else {
			printf( "- " );
			ret = 1;
		}
		if( optind < argc - 1 ) {
			printf( "\n" );
		}
	}
	printf( "\n" );
	return ret;
}

int show_process_mem( int argc, char** argv, int optind )
{
	int _size = 0;
	int _res = 0;
	int _shared = 0;
	int _text = 0;
	int _shared_libs = 0;
	int _stack = 0;
	int _dirty_pages = 0;
	int ret = 0;
	int pid;
	int r;
	for( ; optind < argc; ++optind ) {
		pid = atoi( argv[ optind ] );
		r = read_mem( pid, &_size, &_res, &_shared, &_text, &_shared_libs, &_stack, &_dirty_pages );
		if( ! r ) {
			if( _show_size ) {
				ret = _size;
			} else if( _show_shared ) {
				ret = _shared;
			} else {
				ret = _res;
			}
			ret = format_mem(ret);
			printf( "%d ", ret );
		} else {
			printf( "- " );
			ret = 1;
		}
	}
	printf( "\n" );
	return ret;
}

void gnuplot(int argc, char** argv, int optind) {
	int c = 0, i;
	FILE* f = fopen(PLOT_CMD, "w");
	if (!f) {
		fprintf(stderr, "could not open file %s for writing\n", PLOT_CMD);
		exit(1);
	}
	fprintf(f, 
		"set terminal postscript eps enhance solid \"Time-Roman\" 18\n"
		"set output \"%s\"\n"
		"set title \"memory usage\"\n"
		"set ylabel \"memory usage\"\n"
		"set xlabel \"tics\"\n",
		PLOT_OUTPUT);
	c = 2;
	for (i = optind; i < argc; ++i, ++c) {
		fprintf(f, "plot \"%s\" using 1:%d title \"%s\" with linespoints\n", _gnuplot, c, argv[i]);
	}
	fclose(f);
}

int logmem( int argc, char** argv, int optind, int loginterval ) 
{
	int pid, i, c, r, tic = 0;
	struct mem_t mem;

	if (_gnuplot) {
		gnuplot(argc, argv, optind);
	}
	printf("# tic ");
	for (i = optind; i < argc; ++i) {
		printf("%s ", argv[i]);
	}
	printf("\n");

	while (1) {
		c = 0;
		printf("%d ", ++tic);
		for (i = optind; i < argc; ++i) {
			pid = atoi(argv[i]);
			r = read_mem_t( pid, &mem );
			if (!r) {				
				printf("%d ", format_mem(get_mem(&mem)));
			} else {
				printf("0 ");
				c++;
			}
		}
		printf("\n");
		fflush(stdout);
		if (c + optind == argc) {
			break;
		}
		usleep(loginterval * 1000);
	}
	return 0;
}

